"""
This module contains the API classes for :py:mod:`kinisi`.
It is anticipated that this is where the majority of interaction with the package will occur.
This module includes:

* the :py:class:`kinisi.analyze.DiffusionAnalyzer` class for MSD and diffusion analysis;
* the :py:class:`kinisi.analyze.JumpDiffusionAnalyzer` class for TMSD and collective diffusion analysis; and
* the :py:class:`kinisi.analyze.ConductivityAnalyzer` class for MSCD and conductivity analysis.

These are all compatible with VASP Xdatcar output files, pymatgen structures and any MD trajectory that the
:py:mod:`MDAnalysis` package can handle.
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (arm61)

from typing import Union, List
import numpy as np
from kinisi import diffusion
from kinisi.parser import MDAnalysisParser, PymatgenParser


class Analyzer:
    """
    This class is the superclass for the :py:class:`kinisi.analyze.DiffusionAnalyzer`, 
    :py:class:`kinisi.analyze.JumpDiffusionAnalyzer` and :py:class:`kinisi.analyze.ConductivityAnalyzer` classes. 
    Therefore all of the properties here are available to these other classes. 

    :param delta_t: An array of the timestep values.
    :param disp_3d: A list of arrays, where each array has the axes :code:`[atom, displacement observation,
        dimension]`. There is one array in the list for each delta_t value. Note: it is necessary to use a
        list of arrays as the number of observations is not necessary the same at each data point.
    :param volume: The volume of the simulation cell.
    """
    def __init__(self, delta_t: np.ndarray, disp_3d: List[np.ndarray], volume: float):
        self._delta_t = delta_t
        self._disp_3d = disp_3d
        self._volume = volume

    @classmethod
    def _from_pymatgen(cls,
                       trajectory: List[Union['pymatgen.core.structure.Structure',
                                              List['pymatgen.core.structure.Structure']]],
                       parser_params: dict,
                       dtype: str = None,
                       **kwargs):
        """
        Create a :py:class:`Analyzer` object from a list or nested list of
        :py:class:`pymatgen.core.structure.Structure` objects.

        :param trajectory: The list or nested list of structures to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object.
            See the appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a list of :py:class:`pymatgen.core.structure.Structure`
            objects, this should be :py:attr:`None`. However, if a list of lists is passed, then it is necessary
            to identify if these constitute a series of :py:attr:`consecutive` trajectories or a series of
            :py:attr:`identical` starting points with different random seeds, in which case the `dtype` should
            be either :py:attr:`consecutive` or :py:attr:`identical`.

        :return: Relevant :py:class:`Analyzer` object.
        """
        if dtype is None:
            u = PymatgenParser(trajectory, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        elif dtype == 'identical':
            u = [PymatgenParser(f, **parser_params) for f in trajectory]
            return cls(u[0].delta_t, cls._stack_trajectories(u), u[0].volume, **kwargs)
        elif dtype == 'consecutive':
            structures = _flatten_list([x for x in trajectory])
            u = PymatgenParser(structures, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        else:
            raise ValueError('The dtype specified was not recognised, please consult the kinisi documentation.')

    @classmethod
    def _from_Xdatcar(cls,
                      trajectory: Union['pymatgen.io.vasp.outputs.Xdatcar', List['pymatgen.io.vasp.outputs.Xdatcar']],
                      parser_params: dict,
                      dtype: str = None,
                      **kwargs):
        """
        Create a :py:class:`Analyzer` object from a single or a list of
        :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects.

        :param trajectory: The Xdatcar or list of Xdatcar objects to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a :py:class:`pymatgen.io.vasp.outputs.Xdatcar` object, this
            should be :py:attr:`None`. However, if a list of :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects is
            passed, then it is necessary to identify if these constitute a series of :py:attr:`consecutive`
            trajectories or a series of :py:attr:`identical` starting points with different random seeds, in which
            case the `dtype` should be either :py:attr:`consecutive` or :py:attr:`identical`.

        :return: Relevant :py:class:`Analyzer` object.
        """
        if dtype is None:
            structures = trajectory.structures
            u = PymatgenParser(structures, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        elif dtype == 'identical':
            u = [PymatgenParser(f.structures, **parser_params) for f in trajectory]
            return cls(u[0].delta_t, cls._stack_trajectories(u), u[0].volume, **kwargs)
        elif dtype == 'consecutive':
            structures = _flatten_list([x.structures for x in trajectory])
            u = PymatgenParser(structures, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        else:
            raise ValueError('The dtype specified was not recognised, please consult the kinisi documentation.')

    @classmethod
    def _from_file(cls, trajectory: Union[str, List[str]], parser_params: dict, dtype: str = None, **kwargs):
        """
        Create a :py:class:`Analyzer` object from a single or a list of Xdatcar file(s).

        :param trajectory: The file or list of Xdatcar files to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify if these constitute a series of
            :py:attr:`consecutive` trajectories or a series of :py:attr:`identical` starting points with different
            random seeds, in which case the `dtype` should be either :py:attr:`consecutive` or :py:attr:`identical`.

        :return: Relevant :py:class:`Analyzer` object.
        """
        try:
            from pymatgen.io.vasp import Xdatcar
        except ModuleNotFoundError:  # pragma: no cover
            raise ModuleNotFoundError("To use the from_file method, pymatgen must be installed.")  # pragma: no cover
        if dtype is None:
            structures = Xdatcar(trajectory).structures
            u = PymatgenParser(structures, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        elif dtype == 'identical':
            u = [PymatgenParser(Xdatcar(f).structures, **parser_params) for f in trajectory]
            return cls(u[0].delta_t, cls._stack_trajectories(u), u[0].volume, **kwargs)
        elif dtype == 'consecutive':
            trajectory_list = (Xdatcar(f) for f in trajectory)
            structures = _flatten_list([x.structures for x in trajectory_list])
            u = PymatgenParser(structures, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        else:
            raise ValueError('The dtype specified was not recognised, please consult the kinisi documentation.')

    @classmethod
    def _from_universe(cls,
                       trajectory: 'MDAnalysis.core.universe.Universe',
                       parser_params: dict,
                       dtype: str = None,
                       **kwargs):
        """
        Create an :py:class:`Analyzer` object from an :py:class:`MDAnalysis.core.universe.Universe` object.

        :param trajectory: The universe to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.MDAnalysisParser` object.
            See the appropriate documention for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify that these constitute a series of
            :py:attr:`identical` starting points with different random seeds, in which case the `dtype` should
            be :py:attr:`identical`. For a series of consecutive trajectories, please construct the relevant
            object using :py:mod:`MDAnalysis`.

        :return: Relevant :py:class:`Analyzer` object.
        """
        try:
            import MDAnalysis as mda
        except ModuleNotFoundError:  # pragma: no cover
            raise ModuleNotFoundError(
                "To use the MDAnalysis from file parsing, MDAnalysis must be installed.")  # pragma: no cover
        if dtype is None:
            u = MDAnalysisParser(trajectory, **parser_params)
            return cls(u.delta_t, u.disp_3d, u.volume, **kwargs)
        elif dtype == 'identical':
            u = [MDAnalysisParser(t, **parser_params) for t in trajectory]
            return cls(u[0].delta_t, cls._stack_trajectories(u), u[0].volume, **kwargs)
        else:
            raise ValueError('The dtype specified was not recognised, please consult the kinisi documentation.')

    @staticmethod
    def _stack_trajectories(u: Union[MDAnalysisParser, PymatgenParser]) -> List[np.ndarray]:
        """
        If more than one trajectory is given, then they are stacked to give the appearence that there are
        additional atoms in the trajectory.

        :param u: Results from the parsing of each trajectory.

        :return: The stacked displacement list.
        """
        joint_disp_3d = []
        for i in range(len(u[0].disp_3d)):
            disp = np.zeros((u[0].disp_3d[i].shape[0] * len(u), u[0].disp_3d[i].shape[1], u[0].disp_3d[i].shape[2]))
            disp[:u[0].disp_3d[i].shape[0]] = u[0].disp_3d[i]
            for j in range(1, len(u)):
                disp[u[0].disp_3d[i].shape[0] * j:u[0].disp_3d[i].shape[0] * (j + 1)] = u[j].disp_3d[i]
            joint_disp_3d.append(disp)
        return joint_disp_3d

    @property
    def dt(self) -> np.ndarray:
        """
        :return: Timestep values that have been sampled.
        """
        return self._diff.dt

    @property
    def dr(self) -> List['uravu.distribution.Distribution']:
        """
        :return: A list of :py:class:`uravu.distribution.Distribution` objects that describe the euclidian
            displacement at each :py:attr:`dt`.
        """
        return self._diff.euclidian_displacements

    @property
    def ngp_max(self) -> float:
        """
        :return: Position in dt where the non-Gaussian parameter is maximised.
        """
        return self.dt[self._diff.ngp.argmax()]

    @property
    def intercept(self) -> 'uravu.distribution.Distribution':
        """
        :return: The distribution describing the intercept.
        """
        return self._diff.intercept

    @property
    def volume(self) -> float:
        """
        :return: Volume of system, in cubic angstrom.
        """
        return self._volume


class DiffusionAnalyzer(Analyzer):
    """
    The :py:class:`kinisi.analyze.DiffusionAnalyzer` class performs analysis of diffusion relationships in
    materials.
    This is achieved through the application of a bootstrapping methodology to obtain the most statistically
    accurate values for mean squared displacement uncertainty and estimating the covariance.
    The time-dependence of the MSD is then modelled in a generalised least squares fashion to obtain the diffusion
    coefficient and offset using Markov chain Monte Carlo maximum likelihood sampling.

    :param delta_t: An array of the timestep values.
    :param disp_3d: A list of arrays, where each array has the axes [atom, displacement observation, dimension].
        There is one array in the list for each delta_t value. Note: it is necessary to use a list of arrays as
        the number of observations is not necessary the same at each data point.
    :param volume: The volume of the simulation cell.
    :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.DiffBootstrap` object. See
        the appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.
    """
    def __init__(self,
                 delta_t: np.ndarray,
                 disp_3d: List[np.ndarray],
                 volume: float,
                 bootstrap_params: Union[dict, None] = None):
        if bootstrap_params is None:
            bootstrap_params = {}
        super().__init__(delta_t, disp_3d, volume)
        self._diff = diffusion.MSDBootstrap(self._delta_t, self._disp_3d, **bootstrap_params)

    @classmethod
    def from_pymatgen(cls,
                      trajectory: List[Union['pymatgen.core.structure.Structure',
                                             List['pymatgen.core.structure.Structure']]],
                      parser_params: dict,
                      dtype: str = None,
                      bootstrap_params: dict = None):
        """
        Create a :py:class:`DiffusionAnalyzer` object from a list or nested list of
        :py:class:`pymatgen.core.structure.Structure` objects.

        :param trajectory: The list or nested list of structures to be analysed.
        :dtype trajectory: List[Union['pymatgen.core.structure.Structure', List['pymatgen.core.structure.Structure']]]
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a list of :py:class:`pymatgen.core.structure.Structure` objects,
            this should be :py:attr:`None`. However, if a list of lists is passed, then it is necessary to identify if
            these constitute a series of :py:attr:`consecutive` trajectories or a series of :py:attr:`identical`
            starting points with different random seeds, in which case the `dtype` should be either
            :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`DiffusionAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params kwarg.
        return super()._from_pymatgen(trajectory, parser_params, dtype=dtype, bootstrap_params=bootstrap_params)

    @classmethod
    def from_Xdatcar(cls,
                     trajectory: Union['pymatgen.io.vasp.outputs.Xdatcar', List['pymatgen.io.vasp.outputs.Xdatcar']],
                     parser_params: dict,
                     dtype: str = None,
                     bootstrap_params: dict = None):
        """
        Create a :py:class:`DiffusionAnalyzer` object from a single or a list of
        :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects.

        :param trajectory: The Xdatcar or list of Xdatcar objects to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a :py:class:`pymatgen.io.vasp.outputs.Xdatcar` object, this should
            be :py:attr:`None`. However, if a list of :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects is passed,
            then it is necessary to identify if these constitute a series of :py:attr:`consecutive` trajectories or a
            series of :py:attr:`identical` starting points with different random seeds, in which case the `dtype`
            should be either :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`DiffusionAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params kwarg.
        return super()._from_Xdatcar(trajectory, parser_params, dtype=dtype, bootstrap_params=bootstrap_params)

    @classmethod
    def from_file(cls,
                  trajectory: Union[str, List[str]],
                  parser_params: dict,
                  dtype: str = None,
                  bootstrap_params: dict = None):
        """
        Create a :py:class:`DiffusionAnalyzer` object from a single or a list of Xdatcar file(s).

        :param trajectory: The file or list of Xdatcar files to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify if these constitute a series of
            :py:attr:`consecutive` trajectories or a series of :py:attr:`identical` starting points with different
            random seeds, in which case the `dtype` should be either :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`DiffusionAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params kwarg.
        return super()._from_file(trajectory, parser_params, dtype=dtype, bootstrap_params=bootstrap_params)

    @classmethod
    def from_universe(cls,
                      trajectory: 'MDAnalysis.core.universe.Universe',
                      parser_params: dict,
                      dtype: str = None,
                      bootstrap_params: dict = None):
        """
        Create an :py:class:`DiffusionAnalyzer` object from an :py:class:`MDAnalysis.core.universe.Universe` object.

        :param trajectory: The universe to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.MDAnalysisParser` object.
            See the appropriate documention for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify that these constitute a series of
            :py:attr:`identical` starting points with different random seeds, in which case the `dtype` should
            be :py:attr:`identical`. For a series of consecutive trajectories, please construct the relevant
            object using :py:mod:`MDAnalysis`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`DiffusionAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params kwarg.
        return super()._from_universe(trajectory, parser_params, dtype=dtype, bootstrap_params=bootstrap_params)

    def diffusion(self, diffusion_params: Union[dict, None] = None):
        """
        Calculate the diffusion coefficicent using the bootstrap-GLS methodology.

        :param diffusion_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object.
            See the appropriate documentation for more guidance on this. Optional, default is the default bootstrap
            parameters.
        """
        if diffusion_params is None:
            diffusion_params = {}
        self._diff.diffusion(**diffusion_params)

    @property
    def msd(self) -> np.ndarray:
        """
        :return: MSD for the input trajectories. Note that this is the bootstrap sampled MSD, not the numerical
        average from the data.
        """
        return self._diff.n

    @property
    def msd_std(self) -> np.ndarray:
        """
        :return: MSD standard deviations values for the input trajectories.
        """
        return self._diff.s

    @property
    def D(self) -> 'uravu.distribution.Distribution':
        """
        :return: Diffusion coefficient distribution.
        """
        return self._diff.D

    @property
    def flatchain(self) -> np.ndarray:
        """
        :return: sampling flatchain
        """
        return np.array([self.D.samples, self.intercept.samples]).T


class JumpDiffusionAnalyzer(Analyzer):
    """
    The :py:class:`kinisi.analyze.JumpDiffusionAnalyzer` class performs analysis of collective diffusion
    relationships in materials.
    This is achieved through the application of a bootstrapping methodology to obtain the most statistically
    accurate values for total mean squared displacement uncertainty and covariance.
    The time-dependence of the TMSD is then modelled in a generalised least squares fashion to obtain the jump
    diffusion coefficient and offset using Markov chain Monte Carlo maximum likelihood sampling.

    :param delta_t: An array of the timestep values.
    :param disp_3d: A list of arrays, where each array has the axes [atom, displacement observation, dimension].
        There is one array in the list for each delta_t value. Note: it is necessary to use a list of arrays as
        the number of observations is not necessary the same at each data point.
    :param volume: The volume of the simulation cell.
    :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.DiffBootstrap` object. See
        the appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.
    """
    def __init__(self,
                 delta_t: np.ndarray,
                 disp_3d: List[np.ndarray],
                 volume: float,
                 bootstrap_params: Union[dict, None] = None):
        if bootstrap_params is None:
            bootstrap_params = {}
        super().__init__(delta_t, disp_3d, volume)
        self._diff = diffusion.TMSDBootstrap(self._delta_t, self._disp_3d, **bootstrap_params)

    @classmethod
    def from_pymatgen(cls,
                      trajectory: List[Union['pymatgen.core.structure.Structure',
                                             List['pymatgen.core.structure.Structure']]],
                      parser_params: dict,
                      dtype: str = None,
                      bootstrap_params: dict = None):
        """
        Create a :py:class:`JumpDiffusionAnalyzer` object from a list or nested list of
        :py:class:`pymatgen.core.structure.Structure` objects.

        :param trajectory: The list or nested list of structures to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a list of :py:class:`pymatgen.core.structure.Structure` objects,
            this should be :py:attr:`None`. However, if a list of lists is passed, then it is necessary to identify if
            these constitute a series of :py:attr:`consecutive` trajectories or a series of :py:attr:`identical`
            starting points with different random seeds, in which case the `dtype` should be either
            :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`JumpDiffusionAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params kwarg.
        return super()._from_pymatgen(trajectory, parser_params, dtype=dtype, bootstrap_params=bootstrap_params)

    @classmethod
    def from_Xdatcar(cls,
                     trajectory: Union['pymatgen.io.vasp.outputs.Xdatcar', List['pymatgen.io.vasp.outputs.Xdatcar']],
                     parser_params: dict,
                     dtype: str = None,
                     bootstrap_params: dict = None):
        """
        Create a :py:class:`JumpDiffusionAnalyzer` object from a single or a list of
        :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects.

        :param trajectory: The Xdatcar or list of Xdatcar objects to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a :py:class:`pymatgen.io.vasp.outputs.Xdatcar` object, this should
            be :py:attr:`None`. However, if a list of :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects is passed,
            then it is necessary to identify if these constitute a series of :py:attr:`consecutive` trajectories or a
            series of :py:attr:`identical` starting points with different random seeds, in which case the `dtype`
            should be either :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`JumpDiffusionAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params kwarg.
        return super()._from_Xdatcar(trajectory, parser_params, dtype=dtype, bootstrap_params=bootstrap_params)

    @classmethod
    def from_file(cls,
                  trajectory: Union[str, List[str]],
                  parser_params: dict,
                  dtype: str = None,
                  bootstrap_params: dict = None):
        """
        Create a :py:class:`JumpDiffusionAnalyzer` object from a single or a list of Xdatcar file(s).

        :param trajectory: The file or list of Xdatcar files to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify if these constitute a series of
            :py:attr:`consecutive` trajectories or a series of :py:attr:`identical` starting points with different
            random seeds, in which case the `dtype` should be either :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`JumpDiffusionAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params kwarg.
        return super()._from_file(trajectory, parser_params, dtype=dtype, bootstrap_params=bootstrap_params)

    @classmethod
    def from_universe(cls,
                      trajectory: 'MDAnalysis.core.universe.Universe',
                      parser_params: dict,
                      dtype: str = None,
                      bootstrap_params: dict = None):
        """
        Create an :py:class:`JumpDiffusionAnalyzer` object from an :py:class:`MDAnalysis.core.universe.Universe` object.

        :param trajectory: The universe to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.MDAnalysisParser` object.
            See the appropriate documention for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify that these constitute a series of
            :py:attr:`identical` starting points with different random seeds, in which case the `dtype` should
            be :py:attr:`identical`. For a series of consecutive trajectories, please construct the relevant
            object using :py:mod:`MDAnalysis`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.

        :return: Relevant :py:class:`JumpDiffusionAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params kwarg.
        return super()._from_universe(trajectory, parser_params, dtype=dtype, bootstrap_params=bootstrap_params)

    def jump_diffusion(self, jump_diffusion_params: Union[dict, None] = None):
        """
        Calculate the jump diffusion coefficicent using the bootstrap-GLS methodology.

        :param ump_diffusion_params: The parameters for the :py:class:`kinisi.diffusion.TMSDBootstrap`
            object. See the appropriate documentation for more guidance on this. Optional, default is the
            default bootstrap parameters.
        """
        if jump_diffusion_params is None:
            jump_diffusion_params = {}
        self._diff.jump_diffusion(**jump_diffusion_params)

    @property
    def tmsd(self) -> np.ndarray:
        """
        :return: TMSD for the input trajectories. Note that this is the bootstrap sampled MSD, not the numerical
            average from the data.
        """
        return self._diff.n

    @property
    def tmsd_std(self) -> np.ndarray:
        """
        :return: MSD standard deviations values for the input trajectories.
        """
        return self._diff.s

    @property
    def D_J(self) -> 'uravu.distribution.Distribution':
        """
        :return: Jump diffusion coefficient
        """
        return self._diff.D_J

    @property
    def flatchain(self) -> np.ndarray:
        """
        :return: sampling flatchain
        """
        return np.array([self.D_J.samples, self.intercept.samples]).T


class ConductivityAnalyzer(Analyzer):
    """
    The :py:class:`kinisi.analyze.ConductivityAnalyzer` class performs analysis of conductive relationships in
    materials.
    This is achieved through the application of a bootstrapping methodology to obtain the most statistically
    accurate values for mean squared charge displacement uncertainty and covariance.
    The time-dependence of the MSCD is then modelled in a generalised least squares fashion to obtain the jump
    diffusion coefficient and offset using Markov chain Monte Carlo maximum likelihood sampling.

    :param delta_t: An array of the timestep values.
    :param disp_3d: A list of arrays, where each array has the axes [atom, displacement observation, dimension].
        There is one array in the list for each delta_t value. Note: it is necessary to use a list of arrays as
        the number of observations is not necessary the same at each data point.
    :param volume: The volume of the simulation cell.
    :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.DiffBootstrap` object. See
        the appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.
    :param ionic_charge: The charge on the mobile ions, either an array with a value for each ion or a scalar
        if all values are the same. Optional, default is :py:attr:`1`.
    """
    def __init__(self,
                 delta_t: np.ndarray,
                 disp_3d: List[np.ndarray],
                 volume: float,
                 bootstrap_params: Union[dict, None] = None,
                 ionic_charge: Union[np.ndarray, int] = 1):
        if bootstrap_params is None:
            bootstrap_params = {}
        super().__init__(delta_t, disp_3d, volume)
        self._diff = diffusion.MSCDBootstrap(self._delta_t, self._disp_3d, ionic_charge, **bootstrap_params)

    @classmethod
    def from_pymatgen(cls,
                      trajectory: List[Union['pymatgen.core.structure.Structure',
                                             List['pymatgen.core.structure.Structure']]],
                      parser_params: dict,
                      dtype: str = None,
                      bootstrap_params: dict = None,
                      ionic_charge: Union[np.ndarray, int] = 1):
        """
        Create a :py:class:`ConductivityAnalyzer` object from a list or nested list of
        :py:class:`pymatgen.core.structure.Structure` objects.

        :param trajectory: The list or nested list of structures to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a list of :py:class:`pymatgen.core.structure.Structure` objects,
            this should be :py:attr:`None`. However, if a list of lists is passed, then it is necessary to identify if
            these constitute a series of :py:attr:`consecutive` trajectories or a series of :py:attr:`identical`
            starting points with different random seeds, in which case the `dtype` should be either
            :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.
        :param ionic_charge: The charge on the mobile ions, either an array with a value for each ion or a scalar
            if all values are the same. Optional, default is :py:attr:`1`.

        :return: Relevant :py:class:`ConductivityAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params kwarg.
        return super()._from_pymatgen(trajectory,
                                      parser_params,
                                      dtype=dtype,
                                      bootstrap_params=bootstrap_params,
                                      ionic_charge=ionic_charge)

    @classmethod
    def from_Xdatcar(cls,
                     trajectory: Union['pymatgen.io.vasp.outputs.Xdatcar', List['pymatgen.io.vasp.outputs.Xdatcar']],
                     parser_params: dict,
                     dtype: str = None,
                     bootstrap_params: dict = None,
                     ionic_charge: Union[np.ndarray, int] = 1):
        """
        Create a :py:class:`ConductivityAnalyzer` object from a single or a list of
        :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects.

        :param trajectory: The Xdatcar or list of Xdatcar objects to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a :py:class:`pymatgen.io.vasp.outputs.Xdatcar` object, this should
            be :py:attr:`None`. However, if a list of :py:class:`pymatgen.io.vasp.outputs.Xdatcar` objects is passed,
            then it is necessary to identify if these constitute a series of :py:attr:`consecutive` trajectories or a
            series of :py:attr:`identical` starting points with different random seeds, in which case the `dtype`
            should be either :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.
        :param ionic_charge: The charge on the mobile ions, either an array with a value for each ion or a scalar
            if all values are the same. Optional, default is :py:attr:`1`.

        :return: Relevant :py:class:`ConductivityAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params kwarg.
        return super()._from_Xdatcar(trajectory,
                                     parser_params,
                                     dtype=dtype,
                                     bootstrap_params=bootstrap_params,
                                     ionic_charge=ionic_charge)

    @classmethod
    def from_file(cls,
                  trajectory: Union[str, List[str]],
                  parser_params: dict,
                  dtype: str = None,
                  bootstrap_params: dict = None,
                  ionic_charge: Union[np.ndarray, int] = 1):
        """
        Create a :py:class:`ConductivityAnalyzer` object from a single or a list of Xdatcar file(s).

        :param trajectory: The file or list of Xdatcar files to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.PymatgenParser` object. See the
            appropriate documentation for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify if these constitute a series of
            :py:attr:`consecutive` trajectories or a series of :py:attr:`identical` starting points with different
            random seeds, in which case the `dtype` should be either :py:attr:`consecutive` or :py:attr:`identical`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.
        :param ionic_charge: The charge on the mobile ions, either an array with a value for each ion or a scalar
            if all values are the same. Optional, default is :py:attr:`1`.

        :return: Relevant :py:class:`ConductivityAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params and ionic_charge kwarg.
        return super()._from_file(trajectory,
                                  parser_params,
                                  dtype=dtype,
                                  bootstrap_params=bootstrap_params,
                                  ionic_charge=ionic_charge)

    @classmethod
    def from_universe(cls,
                      trajectory: 'MDAnalysis.core.universe.Universe',
                      parser_params: dict,
                      dtype: str = None,
                      bootstrap_params: dict = None,
                      ionic_charge: Union[np.ndarray, int] = 1):
        """
        Create an :py:class:`ConductivityAnalyzer` object from an :py:class:`MDAnalysis.core.universe.Universe` object.

        :param trajectory: The universe to be analysed.
        :param parser_params: The parameters for the :py:class:`kinisi.parser.MDAnalysisParser` object.
            See the appropriate documention for more guidance on this dictionary.
        :param dtype: If :py:attr:`trajectory` is a single file, this should be :py:attr:`None`. However, if a
            list of files is passed, then it is necessary to identify that these constitute a series of
            :py:attr:`identical` starting points with different random seeds, in which case the `dtype` should
            be :py:attr:`identical`. For a series of consecutive trajectories, please construct the relevant
            object using :py:mod:`MDAnalysis`.
        :param bootstrap_params: The parameters for the :py:class:`kinisi.diffusion.MSDBootstrap` object. See the
            appropriate documentation for more guidance on this. Optional, default is the default bootstrap parameters.
        :param ionic_charge: The charge on the mobile ions, either an array with a value for each ion or a scalar
            if all values are the same. Optional, default is :py:attr:`1`.

        :return: Relevant :py:class:`ConductivityAnalyzer` object.
        """
        # This exists to offer better documentation, in particular for the boostrap_params kwarg.
        return super()._from_universe(trajectory,
                                      parser_params,
                                      dtype=dtype,
                                      bootstrap_params=bootstrap_params,
                                      ionic_charge=ionic_charge)

    def conductivity(self, temperature: float, conductivity_params: Union[dict, None] = None):
        """
        Calculate the jump diffusion coefficicent using the bootstrap-GLS methodology.

        :param temperature: Simulation temperature in Kelvin
        :param conductivity_params: The parameters for the :py:class:`kinisi.diffusion.MSCDBootstrap` object.
            See the appropriate documentation for more guidance on this. Optional, default is the default
            bootstrap parameters
        """
        if conductivity_params is None:
            conductivity_params = {}
        self._diff.conductivity(temperature, self._volume, **conductivity_params)

    @property
    def mscd(self) -> np.ndarray:
        """
        :return: MSCD for the input trajectories. Note that this is the bootstrap sampled value, not the numerical
            average from the data.
        """
        return self._diff.n

    @property
    def mscd_std(self) -> np.ndarray:
        """
        :return: MSCD standard deviations values for the input trajectories.
        """
        return self._diff.s

    @property
    def sigma(self) -> 'uravu.distribution.Distribution':
        """
        :returns: Conductivity, in mS^{1}cm^{-1}.
        """
        return self._diff.sigma

    @property
    def flatchain(self) -> np.ndarray:
        """
        :return: sampling flatchain
        """
        return np.array([self.sigma.samples, self.intercept.samples]).T


def _flatten_list(this_list: list) -> list:
    """
    Flatten nested lists.

    :param this_list: List to be flattened

    :return: Flattened list
    """
    return [item for sublist in this_list for item in sublist]



"""
A small module to find the nearest positive definite matrix.
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (arm61)

import warnings
import numpy as np


def find_nearest_positive_definite(matrix: np.ndarray) -> np.ndarray:
    """
    Find the nearest positive-definite matrix to that given, using the method from N.J. Higham, "Computing a nearest
    symmetric positive semidefinite matrix" (1988): 10.1016/0024-3795(88)90223-6

    :param matrix: Matrix to find nearest positive-definite for.
    :return: Nearest positive-definite matrix.
    """

    if check_positive_definite(matrix):
        return matrix

    warnings.warn("The estimated covariance matrix was not positive definite, the nearest positive "
                  "definite matrix has been found and will be used.")

    B = (matrix + matrix.T) / 2
    _, s, V = np.linalg.svd(B)
    H = np.dot(V.T, np.dot(np.diag(s), V))
    A2 = (B + H) / 2
    A3 = (A2 + A2.T) / 2
    inv_eps = 0.01 / (np.finfo(A3.dtype).eps)

    if (check_positive_definite(A3) and np.linalg.cond(A3) < inv_eps):
        return A3

    spacing = np.spacing(np.linalg.norm(matrix))
    eye = np.eye(matrix.shape[0])
    k = 1
    while not check_positive_definite(A3) or np.linalg.cond(A3) > inv_eps:
        mineig = np.min(np.real(np.linalg.eigvals(A3)))
        A3 += eye * (-mineig * k**2 + spacing)
        k += 1

    return A3


def check_positive_definite(matrix: np.ndarray) -> bool:
    """
    Checks if a matrix is positive-definite via Cholesky decomposition.

    :param matrix: Matrix to check.
    :return: True for a positive-definite matrix.
    """
    try:
        _ = np.linalg.cholesky(matrix)
        return True
    except np.linalg.LinAlgError:
        return False


"""
The two classes herein enable the determination of the activation energy from a diffusion-Arrhenius or
-Super-Arrhenius plot.
This includes the uncertainty on the activation energy from the MCMC sampling of the plot, with uncertainties on
diffusion.
Furthermore, the classes are build on the Relationship subclass, therefore it is easy to determine the Bayesian
evidence for each of the models with the given data, enabling the distinction between Arrhenius and Super-Arrhenius
diffusion.
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (arm61)

from typing import List, Union, Tuple
import numpy as np
from uravu.relationship import Relationship
from scipy.constants import R, N_A, eV

# Convert R to eV
R_no_mol = R / N_A
R_eV = R_no_mol / eV


class StandardArrhenius(Relationship):
    """
    Evaluate the data with a standard Arrhenius relationship.
    For attributes associated with the :py:class:`uravu.relationship.Relationship` class see that documentation.

    Args:
        temperature (:py:attr:`array_like`): Temperature data in kelvin.
        diffusion (:py:attr:`array_like`): Diffusion coefficient data in cm^2s^{-1}.
        bounds (:py:attr:`tuple`): The minimum and maximum values for each parameters. Defaults to :py:attr:`((0, 1), (0, 1e20))`.
        diffusion_error (:py:attr:`array_like`): Uncertainty in the diffusion coefficient data. Not necessary
            if :py:attr:`diffusion` is :py:attr:`list` of :py:class:`uravu.distribution.Distribution` objects.
        ci_points (:py:attr:`array_like`, optional): The two percentiles at which confidence intervals should be
            found for the variables. Default is :py:attr:`[2.5, 97.5]` (a 95 % confidence interval).
    """
    def __init__(self,
                 temperature: np.ndarray,
                 diffusion: Union[List['uravu.distribution.Distribution'], np.ndarray],
                 bounds: Tuple[Tuple[float, float], Tuple[float, float]] = ((0, 1), (0, 1e20)),
                 diffusion_error: np.ndarray = None,
                 ci_points: Tuple[float, float] = None):
        super().__init__(arrhenius, temperature, diffusion, bounds, ordinate_error=diffusion_error, ci_points=None)

    @property
    def activation_energy(self) -> 'uravu.distribution.Distribution':
        """
        :return: Activated energy distribution in electronvolt.
        """
        return self.variables[0]

    @property
    def preexponential_factor(self) -> 'uravu.distribution.Distribution':
        """
        :return: Preexponential factor.
        """
        return self.variables[1]


def arrhenius(abscissa: np.ndarray, activation_energy: float, prefactor: float) -> np.ndarray:
    """
    Determine the diffusion coefficient for a given activation energy, and prefactor according to the Arrhenius
    equation.

    Args:
        abscissa (:py:attr:`array_like`): The abscissa data.
        activation_energy (:py:attr:`float`): The activation_energy value.
        prefactor (:py:attr:`float`): The prefactor value.

    :return: The diffusion coefficient data.
    """
    return prefactor * np.exp(-1 * activation_energy / (R_eV * abscissa))


class SuperArrhenius(Relationship):
    """
    Evaluate the data with a super-Arrhenius relationship. For attributes associated with the
    :py:class:`uravu.relationship.Relationship` class see that documentation.

    Args:
        temperature (:py:attr:`array_like`): Temperature data in kelvin.
        diffusion (:py:attr:`array_like`): Diffusion coefficient data in cm^2s^{-1}.
        bounds (:py:attr:`tuple`): The minimum and maximum values for each parameters. Defaults to 
            :py:attr:`[(0, 1), (0, 1e20), (0, temperature[0])]`.
        diffusion_error (:py:attr:`array_like`): Uncertainty in the diffusion coefficient data. Not necessary
            if :py:attr:`diffusion` is :py:attr:`list` of :py:class:`uravu.distribution.Distribution` objects.
        ci_points (:py:attr:`array_like`, optional): The two percentiles at which confidence intervals should be
            found for the variables. Default is :py:attr:`[2.5, 97.5]` (a 95 % confidence interval).
    """
    def __init__(self,
                 temperature: np.ndarray,
                 diffusion: Union[List['uravu.distribution.Distribution'], np.ndarray],
                 bounds: Tuple[Tuple[float, float], Tuple[float, float], Tuple[float, float]] = [(0, 1), (0, 1e20),
                                                                                                 (0, None)],
                 diffusion_error: np.ndarray = None,
                 ci_points: Tuple[float, float] = None):
        if bounds[2][1] is None:
            bounds[2] = (0, temperature[0])
        super().__init__(super_arrhenius,
                         temperature,
                         diffusion,
                         bounds,
                         ordinate_error=diffusion_error,
                         ci_points=None)

    @property
    def activation_energy(self) -> 'uravu.distribution.Distribution':
        """
        :return: Activated energy distribution in electronvolt.
        """
        return self.variables[0]

    @property
    def preexponential_factor(self) -> 'uravu.distribution.Distribution':
        """
        :return: Preexponential factor.
        """
        return self.variables[1]

    @property
    def T0(self) -> 'uravu.distribution.Distribution':
        """
        :return: Temperature factor for the VTF equation in kelvin.
        """
        return self.variables[2]


def super_arrhenius(abscissa: np.ndarray, activation_energy: float, prefactor: float, t_zero: float) -> np.ndarray:
    """
    Determine the rate constant for a given activation energy, prefactor, and t_zero according to the
    Vogel–Tammann–Fulcher equation.

    :param abscissa: The abscissa data.
    :param activation_energy: The activation_energy value.
    :param prefactor: The prefactor value.
    :param t_zero: The T_0 value.

    :return: The diffusion coefficient data.
    """
    return prefactor * np.exp(-1 * activation_energy / (R_eV * (abscissa - t_zero)))



"""
Tests for parser module
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (arm61)
# pylint: disable=R0201

import unittest
import numpy as np
import MDAnalysis as mda
from numpy.testing import assert_almost_equal, assert_equal
from pymatgen.io.vasp import Xdatcar
import os
import kinisi
from kinisi import parser

dc = np.random.random(size=(100, 100, 3))
indices = np.arange(0, 100, 1, dtype=int)
time_step = 1.0
step_skip = 1


class TestParser(unittest.TestCase):
    """
    Unit tests for parser module
    """
    def test_parser_init_timestep(self):
        p = parser.Parser(dc, indices, [], time_step, step_skip, min_dt=20)
        assert_equal(p.time_step, time_step)

    def test_parser_init_stepskip(self):
        p = parser.Parser(dc, indices, [], time_step, step_skip, min_dt=20)
        assert_equal(p.step_skip, step_skip)

    def test_parser_init_indices(self):
        p = parser.Parser(dc, indices, [], time_step, step_skip, min_dt=20)
        assert_equal(p.indices, indices)

    def test_parser_init_min_dt(self):
        p = parser.Parser(dc, indices, [], time_step, step_skip, min_dt=20)
        assert_equal(p.min_dt, 20)

    def test_parser_delta_t(self):
        p = parser.Parser(dc, indices, [], time_step, step_skip, min_dt=20)
        assert_equal(p.delta_t.size, 80)

    def test_parser_disp_3d(self):
        p = parser.Parser(dc, indices, [], time_step, step_skip, min_dt=20)
        assert_equal(len(p.disp_3d), 80)
        for i, d in enumerate(p.disp_3d):
            assert_equal(d.shape[0], 100)
            assert_equal(d.shape[1], 80 - i)
            assert_equal(d.shape[2], 3)

    def test_smoothed_timesteps(self):
        p = parser.Parser(dc, indices, [], time_step, step_skip, min_dt=20)
        timesteps = p.smoothed_timesteps(100, 30, indices)
        assert_equal(timesteps, np.arange(20, 100, 1))

    def test_smoothed_timesteps_not_enough(self):
        with self.assertRaises(ValueError):
            p = parser.Parser(dc, indices, [], time_step, step_skip, min_dt=120)
            p.smoothed_timesteps(100, 20, indices)

    def test_smoothed_timesteps_min_dt_zero(self):
        p = parser.Parser(dc, indices, [], time_step, step_skip, min_dt=0)
        timesteps = p.smoothed_timesteps(100, 30, indices)
        assert_equal(timesteps, np.arange(1, 100, 1))

    def test_correct_drift_no_framework(self):
        corrected = parser.Parser.correct_drift([], dc)
        assert_equal(len(corrected), 100)
        for i, d in enumerate(corrected):
            assert_equal(d.shape[0], 100)
            assert_equal(d.shape[1], 3)

    def test_correct_drift_framework(self):
        corrected = parser.Parser.correct_drift([], dc)
        assert_equal(len(corrected), 100)
        for i, d in enumerate(corrected):
            assert_equal(d.shape[0], 100)
            assert_equal(d.shape[1], 3)

    def test_get_disps(self):
        p = parser.Parser(dc, indices, [], time_step, step_skip, min_dt=20)
        dt, disp_3d = p.get_disps(np.arange(20, 100, 1), dc)
        assert_equal(dt, np.arange(20, 100, 1))
        assert_equal(len(disp_3d), 80)
        for i, d in enumerate(disp_3d):
            assert_equal(d.shape[0], 100)
            assert_equal(d.shape[1], 80 - i)
            assert_equal(d.shape[2], 3)

    def test_pymatgen_init(self):
        xd = Xdatcar(os.path.join(os.path.dirname(kinisi.__file__), 'tests/inputs/example_XDATCAR.gz'))
        da_params = {'specie': 'Li', 'time_step': 2.0, 'step_skip': 50, 'min_obs': 50}
        data = parser.PymatgenParser(xd.structures, **da_params)
        assert_almost_equal(data.time_step, 2.0)
        assert_almost_equal(data.step_skip, 50)
        assert_equal(data.indices, list(range(xd.natoms[0])))

    def test_pymatgen_big_timestep(self):
        xd = Xdatcar(os.path.join(os.path.dirname(kinisi.__file__), 'tests/inputs/example_XDATCAR.gz'))
        da_params = {'specie': 'Li', 'time_step': 20.0, 'step_skip': 100, 'min_obs': 50}
        data = parser.PymatgenParser(xd.structures, **da_params)
        assert_almost_equal(data.time_step, 20.0)
        assert_almost_equal(data.step_skip, 100)
        assert_equal(data.indices, list(range(xd.natoms[0])))

    def test_mda_init(self):
        xd = mda.Universe(os.path.join(os.path.dirname(kinisi.__file__), 'tests/inputs/example_LAMMPS.data'),
                          os.path.join(os.path.dirname(kinisi.__file__), 'tests/inputs/example_LAMMPS.dcd'),
                          format='LAMMPS')
        da_params = {'specie': '1', 'time_step': 0.005, 'step_skip': 250, 'min_obs': 50}
        data = parser.MDAnalysisParser(xd, **da_params)
        assert_almost_equal(data.time_step, 0.005)
        assert_almost_equal(data.step_skip, 250)
        assert_equal(data.indices, list(range(204)))

    def test_get_matrix(self):
        matrix = parser._get_matrix([10, 10, 10, 90, 90, 90])
        assert_almost_equal(matrix, np.diag((10, 10, 10)))



"""
Parser functions, including implementation for :py:mod:`pymatgen` compatible VASP files and :py:mod:`MDAnalysis`
compatible trajectories.

This parser borrows heavily from the :py:class`pymatgen.analysis.diffusion_analyzer.DiffusionAnalyzer` class, 
originally authored by Will Richards (wrichard@mit.edu) and Shyue Ping Ong.
We include this statement to note that we make no claim to authorship of that code and make no attack on the
original authors.

In fact, we love pymatgen!
"""

# Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (arm61)

from typing import List, Tuple, Union
import numpy as np
from tqdm import tqdm


class Parser:
    """
    The base class for parsing.

    :param disp: Displacements of atoms with the shape [site, time step, axis].
    :param indices: Indices for the atoms in the trajectory used in the diffusion calculation.
    :param framework_indices: Indices for the atoms in the trajectory that should not be used in the diffusion
        calculation.
    :param time_step: Time step, in picoseconds, between steps in trajectory.
    :param step_skip: Sampling freqency of the trajectory (time_step is multiplied by this number to get the real
        time between output from the simulation file).
    :param min_obs: Minimum number of observations of an atom before including it in the MSD vs dt calculation.
        E.g. If a structure has 10 diffusing atoms, and :py:attr:`min_obs=30`, the MSD vs dt will be calculated up
        to :py:attr:`dt = total_run_time / 3`, so that each diffusing atom is measured at least 3 uncorrelated times.
        Optional, defaults to :py:attr:`30`.
    :param min_dt: Minimum timestep to be evaluated, in the simulation units. Optional, defaults to :py:attr:`100`.
    :param ndelta_t: The number of :py:attr:`delta_t` values to calculate the MSD over. Optional,
        defaults to :py:attr:`75`.
    :param progress: Print progress bars to screen. Optional, defaults to :py:attr:`True`.
    """
    def __init__(self,
                 disp: np.ndarray,
                 indices: np.ndarray,
                 framework_indices: np.ndarray,
                 time_step: float,
                 step_skip: int,
                 min_obs: int = 30,
                 min_dt: int = 1,
                 ndelta_t: int = 75,
                 progress: bool = True):
        self.time_step = time_step
        self.step_skip = step_skip
        self.indices = indices
        self.min_dt = min_dt
        self.ndelta_t = ndelta_t
        self._volume = None

        drift_corrected = self.correct_drift(framework_indices, disp)

        nsteps = drift_corrected.shape[1]

        timesteps = self.smoothed_timesteps(nsteps, min_obs, indices)

        self.delta_t, self.disp_3d = self.get_disps(timesteps, drift_corrected, progress)

    @property
    def volume(self) -> float:
        """
        :return: Volume of system, in cubic angstrom.
        """
        return self._volume

    @staticmethod
    def get_disp(coords: List[np.ndarray], latt: List[np.ndarray]) -> np.ndarray:
        """
        Calculate displacements.

        :param coords: Fractional coordinates for all atoms.
        :param latt: Lattice descriptions.

        :return: Numpy array of with shape [site, time step, axis] describing displacements.
        """
        coords = np.concatenate(coords, axis=1)
        d_coords = coords[:, 1:] - coords[:, :-1]
        d_coords = d_coords - np.round(d_coords)
        f_disp = np.cumsum(d_coords, axis=1)
        c_disp = []
        for i in f_disp:
            c_disp.append([np.dot(d, m) for d, m in zip(i, latt[1:])])
        disp = np.array(c_disp)
        return disp

    @staticmethod
    def correct_drift(framework_indices: np.ndarray, disp: np.ndarray) -> np.ndarray:
        """
        Perform drift correction, such that the displacement is calculated normalised to any framework drift.

        :param framework_indices: Indices for the atoms in the trajectory that should not be used in the diffusion
            calculation.
        :param disp: Numpy array of with shape [site, time step, axis] that describes the displacements.

        :return: Displacements corrected to account for drift of a framework.
        """
        # drift corrected position
        if len(framework_indices) > 0:
            framework_disp = disp[framework_indices]
            drift_corrected = disp - np.average(framework_disp, axis=0)[None, :, :]
        else:
            drift_corrected = disp
        return drift_corrected

    def smoothed_timesteps(self, nsteps: int, min_obs: int, indices: np.ndarray) -> np.ndarray:
        """
        Calculate the smoothed timesteps to be used.

        :param nsteps: Number of time steps.
        :param min_obs: Minimum number of observations to have before including in the MSD vs dt calculation.
            E.g. If a structure has 10 diffusing atoms, and :py:attr:`min_obs=30`, the MSD vs dt will be calculated
            up to :py:attr:`dt = total_run_time / 3`, so that each diffusing atom is measured at least 3
            uncorrelated times.
        :param indices: Indices for the atoms in the trajectory used in the calculation of the diffusion.

        :return: Smoothed timesteps.
        """
        min_dt = int(self.min_dt / (self.step_skip * self.time_step))
        max_dt = min(len(indices) * nsteps // min_obs, nsteps)
        if min_dt == 0:
            min_dt = 1
        if min_dt >= max_dt:
            raise ValueError('Not enough data to calculate diffusivity')
        timesteps = np.arange(min_dt, max_dt, max(int((max_dt - min_dt) / self.ndelta_t), 1))
        return timesteps

    def get_disps(self,
                  timesteps: np.ndarray,
                  drift_corrected: np.ndarray,
                  progress: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """
        Calculate the displacement at each timestep.

        :param timesteps: Smoothed timesteps.
        :param drift_corrected: Drift of framework corrected disp.
        :param progress: Print progress bars to screen. Defaults to :py:attr:`True`.

        :return: Tuple containing: time step intervals and raw displacement.
        """
        delta_t = timesteps * self.time_step * self.step_skip
        disp_3d = []
        if progress:
            iterator = tqdm(timesteps, desc='Getting Displacements')
        else:
            iterator = timesteps
        for timestep in iterator:
            disp = np.subtract(drift_corrected[self.indices, timestep:, :],
                               drift_corrected[self.indices, :-timestep, :])
            disp_3d.append(disp)
        return delta_t, disp_3d


class PymatgenParser(Parser):
    """
    A parser for pymatgen structures.

    :param structures: Structures ordered in sequence of run.
    :param specie: Specie to calculate diffusivity for as a String, e.g. :py:attr:`'Li'`.
    :param time_step: Time step, in picoseconds, between steps in trajectory.
    :param step_skip: Sampling freqency of the trajectory (time_step is multiplied by this number to get the real
        time between output from the simulation file).
    :param min_obs: Minimum number of observations of an atom before including it in the MSD vs dt calculation.
        E.g. If a structure has 10 diffusing atoms, and :py:attr:`min_obs=30`, the MSD vs dt will be calculated up
        to :py:attr:`dt = total_run_time / 3`, so that each diffusing atom is measured at least 3 uncorrelated
        times. Optional, defaults to :py:attr:`30`.
    :param sub_sample_traj: Multiple of the :py:attr:`time_step` to sub sample at. Optional, defaults
        to :py:attr:`1` where all timesteps are used.
    :param min_dt: Minimum timestep to be evaluated, in the simulation units. Optional, defaults to :py:attr:`100`.
    :param ndelta_t: The number of :py:attr:`delta_t` values to calculate the MSD over. Optional,
        defaults to :py:attr:`75`.
    :param progress: Print progress bars to screen. Optional, defaults to :py:attr:`True`.
    """
    def __init__(self,
                 structures: List["pymatgen.core.structure.Structure"],
                 specie: Union["pymatgen.core.periodic_table.Element", "pymatgen.core.periodic_table.Specie"],
                 time_step: float,
                 step_skip: int,
                 min_obs: int = 30,
                 sub_sample_traj: int = 1,
                 min_dt: float = 100,
                 ndelta_t: int = 75,
                 progress: bool = True):
        structure, coords, latt = self.get_structure_coords_latt(structures, sub_sample_traj, progress)

        indices = self.get_indices(structure, specie)

        super().__init__(self.get_disp(coords, latt),
                         indices[0],
                         indices[1],
                         time_step,
                         step_skip,
                         min_obs=min_obs,
                         min_dt=min_dt,
                         ndelta_t=ndelta_t,
                         progress=progress)
        self._volume = structure.volume
        self.delta_t *= 1e-3

    @staticmethod
    def get_structure_coords_latt(
            structures: List["pymatgen.core.structure.Structure"],
            sub_sample_traj: int = 1,
            progress: bool = True) -> Tuple["pymatgen.core.structure.Structure", List[np.ndarray], List[np.ndarray]]:
        """
        Obtain the initial structure and displacement from a :py:attr:`list`
        of :py:class`pymatgen.core.structure.Structure`.

        :param structures: Structures ordered in sequence of run.
        :param sub_sample_traj: Multiple of the :py:attr:`time_step` to sub sample at.
            Optional, default is :py:attr:`1`.
        :param progress: Print progress bars to screen. Optional, defaults to :py:attr:`True`.

        :return: Tuple containing: initial structure, fractional coordinates for all atoms,
            and lattice descriptions.
        """
        coords, latt = [], []
        first = True
        if progress:
            iterator = tqdm(structures[::sub_sample_traj], desc='Reading Trajectory')
        else:
            iterator = structures[::sub_sample_traj]
        for struct in iterator:
            if first:
                structure = struct
                first = False
            coords.append(np.array(struct.frac_coords)[:, None])
            latt.append(struct.lattice.matrix)
        coords.insert(0, coords[0])
        latt.insert(0, latt[0])
        return structure, coords, latt

    @staticmethod
    def get_indices(
        structure: "pymatgen.core.structure.Structure", specie: Union["pymatgen.core.periodic_table.Element",
                                                                      "pymatgen.core.periodic_table.Specie"]
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Determine framework and non-framework indices for a :py:mod:`pymatgen` compatible file.

        :param structure: Initial structure.
        :param specie: Specie to calculate diffusivity for as a String, e.g. :py:attr:`'Li'`.

        :returns: Tuple containing: indices for the atoms in the trajectory used in the calculation of the diffusion
            and indices of framework atoms.
        """
        indices = []
        framework_indices = []
        for i, site in enumerate(structure):
            if site.specie.symbol == specie:
                indices.append(i)
            else:
                framework_indices.append(i)
        return indices, framework_indices


class MDAnalysisParser(Parser):
    """
    A parser that consumes an MDAnalysis.Universe object.

    :param universe: The MDAnalysis object of interest.
    :param specie: Specie to calculate diffusivity for as a String, e.g. :py:attr:`'Li'`.
    :param time_step: Time step, in picoseconds, between steps in trajectory.
    :param step_skip: Sampling freqency of the trajectory (time_step is multiplied by this number to get the real
        time between output from the simulation file).
    :param min_obs: Minimum number of observations of an atom before including it in the MSD vs dt calculation.
        E.g. If a structure has 10 diffusing atoms, and :py:attr:`min_obs=30`, the MSD vs dt will be calculated
        up to :py:attr:`dt = total_run_time / 3`, so that each diffusing atom is measured at least 3 uncorrelated
        times. Optional, defaults to :py:attr:`30`.
    :param sub_sample_atoms: The sampling rate to sample the atoms in the system. Optional, defaults 
        to :py:attr:`1` where all atoms are used.
    :param sub_sample_traj: Multiple of the :py:attr:`time_step` to sub sample at. Optional,
        defaults to :py:attr:`1` where all timesteps are used.
    :param min_dt: Minimum timestep to be evaluated, in the simulation units. Optional,
        defaults to :py:attr:`100`.
    :param ndelta_t: The number of :py:attr:`delta_t` values to calculate the MSD over. Optional,
        defaults to :py:attr:`75`.
    :param progress: Print progress bars to screen. Optional, defaults to :py:attr:`True`.
    """
    def __init__(self,
                 universe: "MDAnalysis.core.universe.Universe",
                 specie: str,
                 time_step: float,
                 step_skip: int,
                 min_obs: int = 30,
                 sub_sample_atoms: int = 1,
                 sub_sample_traj: int = 1,
                 min_dt: float = 100,
                 ndelta_t: int = 75,
                 progress: bool = True):
        structure, coords, latt, volume = self.get_structure_coords_latt(universe, sub_sample_atoms, sub_sample_traj,
                                                                         progress)

        indices = self.get_indices(structure, specie)

        super().__init__(self.get_disp(coords, latt), indices[0], indices[1], time_step, step_skip * sub_sample_traj,
                         min_obs, min_dt, ndelta_t, progress)
        self._volume = volume

    @staticmethod
    def get_structure_coords_latt(
            universe: "MDAnalysis.core.universe.Universe",
            sub_sample_atoms: int = 1,
            sub_sample_traj: int = 1,
            progress: bool = True
    ) -> Tuple["MDAnalysis.core.groups.AtomGroup", List[np.ndarray], List[np.ndarray], float]:
        """
        Obtain the initial structure and displacement from a :py:class:`MDAnalysis.universe.Universe` file.

        :param universe: Universe for analysis.
        :param sub_sample_atoms: Frequency to sub sample the number of atoms. Optional, default is :py:attr:`1`.
        :param sub_sample_traj: Multiple of the :py:attr:`time_step` to sub sample at. Optional,
            default is :py:attr:`1`.
        :param progress: Print progress bars to screen. Optional, defaults to :py:attr:`True`.

        :return: Tuple containing: initial structure, fractional coordinates for all atoms,
            lattice descriptions, and the cell volume
        """
        coords, latt = [], []
        first = True
        if progress:
            iterator = tqdm(universe.trajectory[::sub_sample_traj], desc='Reading Trajectory')
        else:
            iterator = universe.trajectory[::sub_sample_traj]
        volume = 0
        for timestep in iterator:
            if first:
                structure = universe.atoms[::sub_sample_atoms]
                first = False
                volume = timestep.volume
            matrix = _get_matrix(timestep.dimensions)
            inv_matrix = np.linalg.inv(matrix)
            coords.append(np.array(np.dot(universe.atoms[::sub_sample_atoms].positions, inv_matrix))[:, None])
            latt.append(matrix)
        coords.insert(0, coords[0])
        latt.insert(0, latt[0])
        return structure, coords, latt, volume

    @staticmethod
    def get_indices(structure: "MDAnalysis.universe.Universe", specie: str) -> Tuple[np.ndarray, np.ndarray]:
        """
        Determine framework and non-framework indices for an :py:mod:`MDAnalysis` compatible file.

        :param structure: Initial structure.
        :param specie: Specie to calculate diffusivity for as a String, e.g. :py:attr:`'Li'`.

        :return: Tuple containing: indices for the atoms in the trajectory used in the calculation of the
            diffusion and indices of framework atoms.
        """
        indices = []
        framework_indices = []
        if not isinstance(specie, list):
            specie = [specie]
        for i, site in enumerate(structure):
            if site.type in specie:
                indices.append(i)
            else:
                framework_indices.append(i)
        return indices, framework_indices


def _get_matrix(dimensions: np.ndarray) -> np.ndarray:
    """
    Determine the lattice matrix.

    :param dimensions: a, b, c, vectors and alpha, beta, gamma angles.

    :return: Lattice matrix
    """
    angles_r = np.radians(dimensions[3:])
    cos_alpha, cos_beta, cos_gamma = np.cos(angles_r)
    sin_alpha, sin_beta = np.sin(angles_r)[:2]

    val = (cos_alpha * cos_beta - cos_gamma) / (sin_alpha * sin_beta)
    # Sometimes rounding errors result in values slightly > 1.
    val = max(min(val, 1), -1)
    gamma_star = np.arccos(val)

    vector_a = [dimensions[0] * sin_beta, 0.0, dimensions[0] * cos_beta]
    vector_b = [
        -dimensions[1] * sin_alpha * np.cos(gamma_star), dimensions[1] * sin_alpha * np.sin(gamma_star),
        dimensions[1] * cos_alpha
    ]
    vector_c = [0.0, 0.0, float(dimensions[2])]

    return np.array([vector_a, vector_b, vector_c], dtype=np.float64).reshape((3, 3))



"""
Tests for analyze module

Copyright (c) Andrew R. McCluskey and Benjamin J. Morgan

Distributed under the terms of the MIT License

@author: Andrew R. McCluskey
"""

import os
import warnings
import unittest
import numpy as np
from numpy.testing import assert_almost_equal
from pymatgen.io.vasp import Xdatcar
import MDAnalysis as mda
from uravu.distribution import Distribution
import kinisi
from kinisi.analyze import Analyzer, DiffusionAnalyzer, JumpDiffusionAnalyzer, ConductivityAnalyzer, _flatten_list

file_path = os.path.join(os.path.dirname(kinisi.__file__), 'tests/inputs/example_XDATCAR.gz')
xd = Xdatcar(file_path)
da_params = {'specie': 'Li', 'time_step': 2.0, 'step_skip': 50, 'min_obs': 1}
md = mda.Universe(os.path.join(os.path.dirname(kinisi.__file__), 'tests/inputs/example_LAMMPS.data'),
                  os.path.join(os.path.dirname(kinisi.__file__), 'tests/inputs/example_LAMMPS.dcd'),
                  format='LAMMPS')
db_params = {'specie': '1', 'time_step': 0.005, 'step_skip': 250, 'min_obs': 1}


class TestAnalyzer(unittest.TestCase):
    """
    Tests for the Analyzer base class.
    """
    def test_structure_pmg(self):
        a = Analyzer._from_pymatgen(xd.structures, parser_params=da_params)
        assert a._delta_t.size == 139
        assert_almost_equal(a._delta_t.max(), 13.9)
        assert len(a._disp_3d) == 139
        assert a._disp_3d[0].shape == (192, 139, 3)
        assert a._disp_3d[-1].shape == (192, 1, 3)

    def test_xdatcar_pmg(self):
        a = Analyzer._from_Xdatcar(xd, parser_params=da_params)
        assert a._delta_t.size == 139
        assert_almost_equal(a._delta_t.max(), 13.9)
        assert len(a._disp_3d) == 139
        assert a._disp_3d[0].shape == (192, 139, 3)
        assert a._disp_3d[-1].shape == (192, 1, 3)

    def test_file_path_pmg(self):
        a = Analyzer._from_file(file_path, parser_params=da_params)
        assert a._delta_t.size == 139
        assert_almost_equal(a._delta_t.max(), 13.9)
        assert len(a._disp_3d) == 139
        assert a._disp_3d[0].shape == (192, 139, 3)
        assert a._disp_3d[-1].shape == (192, 1, 3)

    def test_identical_structure_pmg(self):
        a = Analyzer._from_pymatgen([xd.structures, xd.structures], parser_params=da_params, dtype='identical')
        assert a._delta_t.size == 139
        assert_almost_equal(a._delta_t.max(), 13.9)
        assert len(a._disp_3d) == 139
        assert a._disp_3d[0].shape == (384, 139, 3)
        assert a._disp_3d[-1].shape == (384, 1, 3)

    def test_identical_xdatcar_pmg(self):
        a = Analyzer._from_Xdatcar([xd, xd], parser_params=da_params, dtype='identical')
        assert a._delta_t.size == 139
        assert_almost_equal(a._delta_t.max(), 13.9)
        assert len(a._disp_3d) == 139
        assert a._disp_3d[0].shape == (384, 139, 3)
        assert a._disp_3d[-1].shape == (384, 1, 3)

    def test_identical_file_path_pmg(self):
        a = Analyzer._from_file([file_path, file_path], parser_params=da_params, dtype='identical')
        assert a._delta_t.size == 139
        assert a._delta_t.max() == 13.90
        assert len(a._disp_3d) == 139
        assert a._disp_3d[0].shape == (384, 139, 3)
        assert a._disp_3d[-1].shape == (384, 1, 3)

    def test_consecutive_structure_pmg(self):
        a = Analyzer._from_pymatgen([xd.structures, xd.structures], parser_params=da_params, dtype='consecutive')
        assert a._delta_t.size == 93
        assert_almost_equal(a._delta_t.max(), 27.7)
        assert len(a._disp_3d) == 93
        assert a._disp_3d[0].shape == (192, 279, 3)
        assert a._disp_3d[-1].shape == (192, 3, 3)

    def test_consecutive_xdatcar_pmg(self):
        a = Analyzer._from_Xdatcar([xd, xd], parser_params=da_params, dtype='consecutive')
        assert a._delta_t.size == 93
        assert_almost_equal(a._delta_t.max(), 27.7)
        assert len(a._disp_3d) == 93
        assert a._disp_3d[0].shape == (192, 279, 3)
        assert a._disp_3d[-1].shape == (192, 3, 3)

    def test_consecutive_file_path_pmg(self):
        a = Analyzer._from_file([file_path, file_path], parser_params=da_params, dtype='consecutive')
        assert a._delta_t.size == 93
        assert_almost_equal(a._delta_t.max(), 27.7)
        assert len(a._disp_3d) == 93
        assert a._disp_3d[0].shape == (192, 279, 3)
        assert a._disp_3d[-1].shape == (192, 3, 3)

    def test_mdauniverse(self):
        a = Analyzer._from_universe(md, parser_params=db_params)
        assert a._delta_t.size == 120
        assert_almost_equal(a._delta_t.max(), 248.75)
        assert len(a._disp_3d) == 120
        assert a._disp_3d[0].shape == (204, 120, 3)
        assert a._disp_3d[-1].shape == (204, 1, 3)

    def test_identical_mdauniverse(self):
        a = Analyzer._from_universe([md, md], parser_params=db_params, dtype='identical')
        assert a._delta_t.size == 120
        assert_almost_equal(a._delta_t.max(), 248.75)
        assert len(a._disp_3d) == 120
        assert a._disp_3d[0].shape == (408, 120, 3)
        assert a._disp_3d[-1].shape == (408, 1, 3)

    def test_list_bad_input(self):
        with self.assertRaises(ValueError):
            a = Analyzer._from_file([file_path, file_path], parser_params=da_params, dtype='consecutie')

    def test_list_bad_mda(self):
        with self.assertRaises(ValueError):
            a = Analyzer._from_universe(file_path, parser_params=db_params, dtype='consecutie')


class TestDiffusionAnalyzer(unittest.TestCase):
    """
    Tests for the DiffusionAnalyzer base class.
    """
    def test_properties(self):
        with warnings.catch_warnings(record=True) as w:
            a = DiffusionAnalyzer.from_pymatgen(xd.structures,
                                                parser_params=da_params,
                                                bootstrap_params={'random_state': np.random.RandomState(0)})
            assert_almost_equal(a.dt, a._diff.dt)
            assert_almost_equal(a.msd, a._diff.n)
            assert_almost_equal(a.msd_std, a._diff.s)
            for i in range(len(a.dr)):
                assert_almost_equal(a.dr[i].samples, a._diff.euclidian_displacements[i].samples)
            assert a.ngp_max == a._diff.dt[a._diff.ngp.argmax()]
            assert a.D == None
            assert issubclass(w[0].category, UserWarning)
            assert "maximum" in str(w[0].message)

    def test_diffusion(self):
        with warnings.catch_warnings(record=True) as w:
            a = DiffusionAnalyzer.from_pymatgen(xd.structures,
                                                parser_params=da_params,
                                                bootstrap_params={'random_state': np.random.RandomState(0)})
            assert_almost_equal(a.dt, a._diff.dt)
            assert_almost_equal(a.msd, a._diff.n)
            assert_almost_equal(a.msd_std, a._diff.s)
            for i in range(len(a.dr)):
                assert_almost_equal(a.dr[i].samples, a._diff.euclidian_displacements[i].samples)
            a.diffusion()
            assert a.ngp_max == a._diff.dt[a._diff.ngp.argmax()]
            assert isinstance(a.D, Distribution)
            assert a.flatchain.shape == (32000, 2)
            assert issubclass(w[0].category, UserWarning)
            assert "maximum" in str(w[0].message)


class TestJumpDiffusionAnalyzer(unittest.TestCase):
    """
    Tests for the JumpDiffusionAnalyzer base class.
    """
    def test_properties(self):
        with warnings.catch_warnings(record=True) as w:
            a = JumpDiffusionAnalyzer.from_pymatgen(xd.structures,
                                                    parser_params=da_params,
                                                    bootstrap_params={'random_state': np.random.RandomState(0)})
            assert_almost_equal(a.dt, a._diff.dt)
            assert_almost_equal(a.tmsd, a._diff.n)
            assert_almost_equal(a.tmsd_std, a._diff.s)
            for i in range(len(a.dr)):
                assert_almost_equal(a.dr[i].samples, a._diff.euclidian_displacements[i].samples)
            assert a.ngp_max == a._diff.dt[a._diff.ngp.argmax()]
            assert issubclass(w[0].category, UserWarning)
            assert "maximum" in str(w[0].message)

    def test_diffusion(self):
        with warnings.catch_warnings(record=True) as w:
            a = JumpDiffusionAnalyzer.from_pymatgen(xd.structures,
                                                    parser_params=da_params,
                                                    bootstrap_params={'random_state': np.random.RandomState(0)})
            assert_almost_equal(a.dt, a._diff.dt)
            assert_almost_equal(a.tmsd, a._diff.n)
            assert_almost_equal(a.tmsd_std, a._diff.s)
            for i in range(len(a.dr)):
                assert_almost_equal(a.dr[i].samples, a._diff.euclidian_displacements[i].samples)
            a.jump_diffusion()
            assert a.ngp_max == a._diff.dt[a._diff.ngp.argmax()]
            assert isinstance(a.D_J, Distribution)
            assert a.flatchain.shape == (32000, 2)
            assert issubclass(w[0].category, UserWarning)
            assert "maximum" in str(w[0].message)


class TestConductivityAnalyzer(unittest.TestCase):
    """
    Tests for the ConductivityAnalyzer base class.
    """
    def test_properties(self):
        with warnings.catch_warnings(record=True) as w:
            a = ConductivityAnalyzer.from_pymatgen(xd.structures,
                                                   parser_params=da_params,
                                                   bootstrap_params={'random_state': np.random.RandomState(0)},
                                                   ionic_charge=1)
            assert_almost_equal(a.dt, a._diff.dt)
            assert_almost_equal(a.mscd, a._diff.n)
            assert_almost_equal(a.mscd_std, a._diff.s)
            for i in range(len(a.dr)):
                assert_almost_equal(a.dr[i].samples, a._diff.euclidian_displacements[i].samples)
            assert a.ngp_max == a._diff.dt[a._diff.ngp.argmax()]
            assert issubclass(w[0].category, UserWarning)
            assert "maximum" in str(w[0].message)

    def test_diffusion(self):
        with warnings.catch_warnings(record=True) as w:
            a = ConductivityAnalyzer.from_pymatgen(xd.structures,
                                                   parser_params=da_params,
                                                   bootstrap_params={'random_state': np.random.RandomState(0)},
                                                   ionic_charge=1)
            assert_almost_equal(a.dt, a._diff.dt)
            assert_almost_equal(a.mscd, a._diff.n)
            assert_almost_equal(a.mscd_std, a._diff.s)
            for i in range(len(a.dr)):
                assert_almost_equal(a.dr[i].samples, a._diff.euclidian_displacements[i].samples)
            a.conductivity(100)
            assert a.ngp_max == a._diff.dt[a._diff.ngp.argmax()]
            assert isinstance(a.sigma, Distribution)
            assert a.flatchain.shape == (32000, 2)
            assert issubclass(w[0].category, UserWarning)
            assert "maximum" in str(w[0].message)


class TestFunctions(unittest.TestCase):
    """
    Tests for other functions
    """
    def test__flatten_list(self):
        a_list = [[1, 2, 3], [4, 5]]
        result = _flatten_list(a_list)
        assert result == [1, 2, 3, 4, 5]
